from __future__ import annotations
from typing import Optional
from direct.task import Task
from panda3d.core import WindowProperties

import roboviz.controller.Scene as S
from abc import ABC, abstractmethod

class Controls:
	schemes: dict[str, ControlScheme]
	__selected: Optional[ControlScheme]
	def __init__(self, schemes: list[ControlScheme], start: Optional[str] = None):
		"""
		`schemes`: A list of classes that extend ControlSchemes to be used by this Controls class

		`start`: (optional) The name of the ControlScheme to start with. Equivalent to calling
		`request(start)`
		"""
		self.schemes = {type(s).__name__: s for s in schemes}
		self.__selected = None
		if start: self.request(start)

	def request(self, name: str) -> bool:
		"""
		Disables the previous ControlScheme and switches to the control scheme with the given `name`
		but DOES NOT ACTIVATE IT. Call `enable` to activate the new scheme.
		If the scheme is already active, this function does nothing. and returns false.
		"""
		assert name in self.schemes, f"Requested Control scheme {name} does not exist"
		if type(self.__selected).__name__ == name: return False
		if self.__selected is not None:
			self.__selected.disable()
			self.__selected.on_switch_away()
		self.__selected = self.schemes[name]
		self.__selected.on_requested()
		return True

	def enable(self) -> None:
		"""Activates the current ControlScheme"""
		assert self.__selected is not None, "No control scheme is selected"
		self.__selected.enable()

	def disable(self) -> None:
		"""Deactivates the current ControlScheme"""
		assert self.__selected is not None, "No control scheme is selected"
		self.__selected.disable()

	def set_zoom(self, zoom: float) -> None:
		assert self.__selected is not None, "No control scheme is selected"
		self.__selected.set_zoom(zoom)

	def set_speed(self, zoom: float) -> None:
		assert self.__selected is not None, "No control scheme is selected"
		self.__selected.set_speed(zoom)

	@property
	def speed(self) -> float:
		"""A number between 0 and 1 representing how fast navigation is in current control scheme"""
		assert self.__selected is not None, "No control scheme is selected"
		return self.__selected.speed

	@property
	def zoom(self) -> float:
		"""
		A number between 0 and 1 which represents how zoomed in the camera is in the current control
		scheme
		"""
		assert self.__selected is not None, "No control scheme is selected"
		return self.__selected.zoom

	@property
	def list(self) -> list[str]:
		"""A list of the class names of all the available control schemes"""
		return [type(i).__name__ for i in self.schemes.values()]

	@property
	def list_names(self) -> list[str]:
		"""A list of the display names of all the available control schemes"""
		return [i.display_name for i in self.schemes.values()]

	@property
	def selected(self) -> str:
		"""The class name of the selected control scheme"""
		assert self.__selected is not None, "No control scheme is selected"
		return type(self.__selected).__name__

	@property
	def description(self) -> str:
		"""The description of the selected control scheme"""
		assert self.__selected is not None, "No control scheme is selected"
		return self.__selected.description

class ControlScheme(ABC):
	"""
	# ControlScheme
	Base class for a panda3d base control scheme for navigating a 3d environment. Meant to be
	extended and used in a Controls object.
	### Required Methods
	`enable(self) -> None` - Called when the control scheme is enabled. This should set self.enabled
	to True and set up the control scheme.

	`disable(self) -> None` - Called when the control scheme is disabled. This should set
	self.enabled to False and clean up the control scheme.

	`on_requested(self) -> None` - Called when the control scheme is requested by the controls
	object. This is where transitions from the previous camera angle should be handled.

	`on_switch_away(self) -> None` - Called when the control scheme is removed from being the active
	control scheme, i.e. when it is switched away from. This is where it should clean up after itself

	`set_zoom(self, zoom: float) -> None` - Called when the zoom level is changed. This should take
	a float between 0 and 1.

	`set_speed(self, speed: float) -> None` - Called when the speed is changed. This should take a
	float between 0 and 1.
	### Required variables
	`display_name: str` - The name of the control scheme displayed to the user

	`description: str` - A description of the controls available in this control scheme

	`is_enabled: bool` - Whether this control scheme is currently active

	`speed: float` - A number between 0 and 1 representing how fast navigation is in the scheme

	`zoom: float` - A number between 0 and 1 which represents how zoomed in the camera is
	"""
	display_name: str
	"""The name of this control scheme displayed to the user"""
	description: str
	"""A description of the controls available in this control scheme"""
	is_enabled: bool
	"""Whether this control scheme is currently active"""
	speed: float
	"""A number between 0 and 1 representing how fast navigation is in the scheme"""
	zoom: float
	"""A number between 0 and 1 which represents how zoomed in the camera is"""

	def __new__(cls):
		obj = super().__new__(cls)
		obj.is_enabled = False
		return obj

	@abstractmethod
	def enable(self) -> None: ...

	@abstractmethod
	def disable(self) -> None: ...

	@abstractmethod
	def on_requested(self) -> None: ...

	@abstractmethod
	def on_switch_away(self) -> None: ...

	@abstractmethod
	def set_zoom(self, zoom: float) -> None: ...

	@abstractmethod
	def set_speed(self, speed: float) -> None: ...

	@property
	def scene(self) -> "S.Scene":
		"""
		When Panda3D creates a base or class that extends base it adds it to builtins. By accessing
		it as a property that is not accessed until the class is instantiated we can avoid a
		circular dependency, error warnings and gain type hinting.
		"""
		return base # type: ignore

class Game(ControlScheme):
	ROTATION_SPEED = 20
	DEFAULT_FOV = 40
	ZOOM_INC = 0.01

	keymap: dict[str, bool]

	def __init__(self):
		super().__init__()
		self.display_name = "Game"
		self.keymap = {
			"w": False,
			"a": False,
			"s": False,
			"d": False,
			"space": False,
			"shift": False,
		}
		self.description = f"""
{'FORWARDS : W':<20}{'BACKWARDS : S':>20}\n
{'LEFT : A':<20}{'RIGHT : D':>20}\n
{'UP : Space':<20}{'DOWN : Shift':>20}\n
{'LOOK : Mouse':<20}{'ZOOM : Scroll':>20}\n
C : Re-Center
		"""
		self.speed = 0.5
		self.zoom = 0.5

	def enable(self) -> None:
		self.is_enabled = True

		props = WindowProperties()	# type: ignore
		props.set_cursor_hidden(True)
		props.setMouseMode(WindowProperties.M_confined)
		self.scene.win.requestProperties(props)	# type: ignore

		self.scene.center_mouse()
		self.scene.taskMgr.add(self.navigate, "navigate")
		self.scene.taskMgr.add(self.look, "look")

		for k in self.keymap:
			self.scene.accept(k, self.update_keymap, [k, True])
			self.scene.accept(f"{k}-up", self.update_keymap, [k, False])

		self.scene.accept("wheel_up", self.adjust_zoom, [Game.ZOOM_INC])
		self.scene.accept("wheel_down", self.adjust_zoom, [-Game.ZOOM_INC])
		self.scene.accept("c", self.scene.camera.look_at, [0, 0, 0])

	def disable(self) -> None:
		self.is_enabled = False
		for k in self.keymap.keys():
			self.keymap[k] = False

		props = WindowProperties() # type: ignore
		props.set_cursor_hidden(False)
		props.set_mouse_mode(WindowProperties.M_absolute)
		self.scene.win.requestProperties(props)  # type: ignore

		self.scene.taskMgr.remove("navigate")
		self.scene.taskMgr.remove("look")
		# ignore all key presses. Other scenes will rebind the keys they need
		self.scene.ignore_all()

	def on_requested(self) -> None:
		self.scene.angle_camera()
		self.scene.task_mgr.add(self.scene.orbit, "orbit")

	def on_switch_away(self) -> None:
		self.scene.task_mgr.remove("orbit")
		self.scene.camLens.set_fov(Game.DEFAULT_FOV)

	def adjust_zoom(self, increment: float) -> None:
		"""Adjust the FOV, pass a positive number to increase, negative to decrease"""
		self.zoom = max(0, min(1, self.zoom + increment))
		self.scene.camLens.set_fov(max(1, self.zoom*Game.DEFAULT_FOV*2))

	def set_zoom(self, zoom: float) -> None:
		self.zoom = zoom
		self.scene.camLens.set_fov(max(1, zoom*Game.DEFAULT_FOV*2))

	def set_speed(self, speed: float) -> None:
		self.speed = speed

	def navigate(self, task: Task.Task) -> int:
		"""Callback to navigate the camera"""
		cam: NodePath = self.scene.camera  # type: ignore
		if (self.keymap["w"]):
			cam.set_y(cam, self.speed)
		if (self.keymap["s"]):
			cam.set_y(cam, -self.speed)
		if (self.keymap["a"]):
			cam.set_x(cam, -self.speed)
		if (self.keymap["d"]):
			cam.set_x(cam, self.speed)
		if (self.keymap["space"]):
			cam.set_z(cam.get_z() + self.speed)
		if (self.keymap["shift"]):
			cam.set_z(cam.get_z() - self.speed)
		return Task.cont

	def look(self, task: Task.Task) -> int:
		"""Callback to look around the scene"""
		mw = self.scene.mouseWatcherNode
		if mw.hasMouse():
			self.scene.camera.set_h(self.scene.camera.get_h() - mw.getMouseX() * Game.ROTATION_SPEED)
			self.scene.camera.set_p(self.scene.camera.get_p() + mw.getMouseY() * Game.ROTATION_SPEED)
			self.scene.center_mouse()

		return Task.cont

	def update_keymap(self, key: str, value: bool) -> None:
		"""Callback to update the keymap"""
		self.keymap[key] = value


class OnlineVisualizer(ControlScheme):
	ROTATION_MULTIPLIER = 50
	PAN_MULTIPLIER = 5

	zoom_multiplier: int
	x_down: float
	y_down: float
	left_down: bool
	right_down: bool
	zoom: float

	def __init__(self):
		super().__init__()
		self.zoom_multiplier = round(self.scene.terrain.get_sx()+self.scene.terrain.get_sy()) * 4
		self.display_name = "Online Visualizer"
		self.description = """
ROTATE : Left Click & Drag\n
PAN : Right Click & Drag\n
ZOOM : Scroll
		"""
		self.left_down = False
		self.right_down = False
		self.zoom = 0.5
		self.speed = 0.5
		# todo implement method that runs when control scheme is requested and put code to set camera position in there to prevent zooming through the floor

	def enable(self) -> None:
		self.is_enabled = True
		self.scene.accept("mouse1", self.on_mouse_down, [True])
		self.scene.accept("mouse1-up", self.on_mouse_up, [True])
		self.scene.accept("mouse3", self.on_mouse_down, [False])
		self.scene.accept("mouse3-up", self.on_mouse_up, [False])
		self.scene.accept("wheel_up", self.adjust_zoom, [0.01])
		self.scene.accept("wheel_down", self.adjust_zoom, [-0.01])
		self.scene.taskMgr.add(self.navigate, "navigate")

	def disable(self) -> None:
		self.is_enabled = False
		self.scene.ignore_all()
		self.left_down = False
		self.scene.taskMgr.remove("navigate")

	def on_requested(self) -> None:
		self.scene.camera.set_pos(self.scene.origin, 0, 0, self.zoom_multiplier/2+5)
		self.scene.camera.look_at(self.scene.origin)

	def on_switch_away(self) -> None:
		self.scene.origin.set_hpr(0, 0, 0)

	def on_mouse_down(self, left: bool) -> None:
		mw = self.scene.mouseWatcherNode
		if mw.hasMouse():
			self.x_down, self.y_down = mw.getMouseX(), mw.getMouseY()
			if left: self.left_down = True
			else: self.right_down = True

	def on_mouse_up(self, left: bool) -> None:
		if left: self.left_down = False
		else: self.right_down = False

	def navigate(self, task: Task.Task) -> int:
		mw = self.scene.mouseWatcherNode
		if not mw.hasMouse(): return Task.cont
		if self.left_down:
			dy = (self.y_down-mw.getMouseY())*OnlineVisualizer.ROTATION_MULTIPLIER
			#get position of scene and split the x and y axis tilts accordingly
			x_rotation = abs(abs(self.scene.origin.get_h()) -90)/90
			y_rotation = 1 - x_rotation
			# ensure terrain always pitches towards the camera
			if abs(self.scene.origin.get_h()) > 90:	x_rotation *= -1
			if self.scene.origin.get_h() > 0:		y_rotation *= -1
			# pitch terrain
			self.scene.origin.set_p(self.scene.origin, x_rotation*dy)
			self.scene.origin.set_r(self.scene.origin, y_rotation*dy)
			# rotate origin
			self.scene.origin.set_h(self.scene.origin, -(self.x_down-mw.getMouseX())*OnlineVisualizer.ROTATION_MULTIPLIER)

		if self.right_down:
			multiplier = OnlineVisualizer.PAN_MULTIPLIER+OnlineVisualizer.PAN_MULTIPLIER**(1+(1-self.zoom))
			self.scene.camera.set_z(self.scene.camera, (self.y_down-mw.getMouseY())*multiplier) # type: ignore
			self.scene.camera.set_x(self.scene.camera, (self.x_down-mw.getMouseX())*multiplier) # type: ignore
		self.y_down = mw.getMouseY()
		self.x_down = mw.getMouseX()
		return Task.cont

	def adjust_zoom(self, quantity: float) -> None:
		self.zoom = max(0, min(1, self.zoom + quantity))
		if 0 < self.zoom and self.zoom < 1:
			self.scene.camera.set_y(self.scene.camera, quantity*self.zoom_multiplier) # type: ignore

	def set_zoom(self, zoom: float) -> None:
		# zoom = 1 - zoom
		self.scene.camera.set_y(self.scene.camera, (zoom - self.zoom)*self.zoom_multiplier) # type: ignore
		self.zoom = zoom


	def set_speed(self, speed: float) -> None:
		pass