from direct.fsm.FSM import FSM as BadFSM, AlreadyInTransition
from direct.gui.DirectGuiBase import DirectGuiWidget
from abc import ABC, abstractmethod
from typing import Any, Callable, Literal, Union


import roboviz.controller.Scene as S # cannot do from Scene import Scene due to circular import
from roboviz.view.Components import Widget

class State(ABC):
	"""
	# State
	Base class for a state in a FSM.

	Extend this class to create a state, and give the sub class the same name as the state.

	### Instance Variable Rules
	DO NOT create any instance variables that start with "to" or "from". This is reserved for
	transition methods.

	### Required methods:
	`enter(self, *args) -> None` - Called when the state is entered. This is where you should create
	your GUI elements.

	`exit(self, *args) -> None` - Called when the state is exited. This is where you should clean up
	the state, destroying GUI elements etc.

	### Optional methods:
	`filter(self, request, args) -> str | None` - Used to determine which state to transition to
	next on the receipt of a particular input. The filter function should return the name of the
	state to transition to, or None if no transition is desired.

	`to<StateName>` - Called instead of exit when this state is transitioning to <StateName>. Used
	for custom transitions between specific states

	`from<StateName>` - Called instead of enter when this state is being transitioned to from
	<StateName>. Used for custom transitions between specific states.
	### The request method
	One of the most common actions in a state is to request the transition to a different state.
	This is done by calling the `request` method on the FSM. This class has a dummy `request` method
	for type hinting purposes. The actual `request` method exists on FSM. Because extensions of this
	class are not used as classes but instead dismembered for use in the FSM it does not matter that
	request does not have any implementation.

	### Utility Methods
	`add_widget(self, widget: Widget)` - Adds a widget to the state. All widgets added in this
	manner will be destroyed when `clear_widgets` is called.

	`clear_widgets(self)` - Destroys all widgets added to the state using `add_widget`. Should be
	called in the `exit` method if any widgets were added.
	"""
	widgets: list[Union[Widget, DirectGuiWidget]]

	def __new__(cls):
		cls = super().__new__(cls)
		cls.widgets = []
		return cls

	@abstractmethod
	def enter(self, *args: Any): ...

	@abstractmethod
	def exit(self, *args: Any): ...

	def request(self, stateName: str, *args: Any) -> (tuple[str, ...] | tuple[Literal['Off'], ...] | Any | None):
		"""
		Requests a state transition (or other behavior). The request may be denied by the FSM's
		filter function. If it is denied, the filter function may either raise an exception
		(`RequestDenied`), or it may simply return None, without changing the FSM's state.

		The request parameter should be a string. The request, along with any additional arguments,
		is passed to the current filterState() function. If filterState() returns a string, the FSM
		transitions to that state.

		The return value is the same as the return value of filterState() (that is, None if the
		request does not provoke a state transition, otherwise it is a tuple containing the name of
		the state followed by any optional args.)

		If the FSM is currently in transition (i.e. in the middle of executing an enterState or
		exitState function), an `AlreadyInTransition` exception is raised (but see `demand()`, which
		will queue these requests up and apply when the transition is complete).
		"""
		...

	def add_widget(self, widget: Union[Widget, DirectGuiWidget]):
		"""
		Add a widget to the state. All widgets added in this manner will be destroyed when
		`clear_widgets` is called.
		"""
		self.widgets.append(widget)

	def clear_widgets(self):
		"""Destroy all widgets added to the state using `add_widget`."""
		for widget in self.widgets:
			widget.destroy()
		self.widgets = []

	@property
	def scene(self) -> "S.Scene":
		"""
		When Panda3D creates a base or class that extends base it adds it to builtins. By accessing
		it as a property that is not accessed until the class is instantiated we can avoid a
		circular dependency, error warnings and gain type hinting.
		"""
		return base # type: ignore


class FSM(BadFSM):
	"""
	A Finite State Machine composed of 0 or more `State` classes passed into the constructor or
	added later with the `addState` method. These states can be transitioned between by calling the
	request method with the name of the state to transition to. The FSM will call the `filter`
	method of the current state to determine if the transition is allowed. If it is the exit method
	of the current state will be called, then the enter method of the new state will be called,
	unless the current state has a method named `to<NewStateName>` in which case that method will be
	called instead of exit.

	This class extends the FSM class from Panda3d's direct.fsm.FSM module in order to allow states
	to be defined as classes instead being implicitly defined by adding enter<StateName> and
	exit<StateName> methods to the FSM class, allowing for far greater modularity and scalability.

	Unfortunately the methods that needed to be overridden were class-private so they could not be
	overridden in the standard way, and I had to resort to private name mangling
	(see https://docs.python.org/2/reference/expressions.html#atom-identifiers). Why the Panda3D
	developers decided to design their FSM class in this way is beyond me.
	"""
	states: dict[str, State]

	# Implementing the singleton design pattern to match the way Panda3d has a global showBase
	_instance = None
	def __new__(cls, *args: Any):
		if cls._instance is None:
			cls._instance = super().__new__(cls)
		return cls._instance

	def __init__(self, name: str, states: list[State]):
		super().__init__(name)
		self.states = {}
		for state in states: self.add_state(state)

	def add_state(self, state: State) -> None:
		setattr(state, "request", self.request)
		name = type(state).__name__
		assert name[0].isupper(), "State name must start with a capital letter"
		assert hasattr(state, "enter"), f"State {name} does not have an enter method"
		assert hasattr(state, "exit"), f"State {name} does not have an exit method"
		self.states[name] = state

	def getCurrentFilter(self) -> Callable:
		if not self.state:
			error = f"FSM cannot determine current filter while in transition ({self.oldState} -> {self.newState})."
			raise AlreadyInTransition(error)

		# If there's no matching filterState() function, call defaultFilter() instead.
		return getattr(self.states.get(self.state, None), "filter", self.defaultFilter)

	def getEnterFunc(self, name: str, *args: Any) -> Callable:
		"""Returns the appropriate enter function when transitioning into a new state, if it exists."""
		assert self.state == None and self.newState == name

		# If there's no matching enter() function, return defaultEnter() instead.
		return getattr(self.states.get(name, None), "enter", self.defaultEnter)

	def _FSM__callFromToFunc(self, oldState: str, newState: str, *args: Any) -> bool:
		"""
		Calls the appropriate from and to functions when transitioning into a new state if they
		exist, or the default enter and exit functions if they don't.
		"""
		assert self.state == None and self.oldState == oldState and self.newState == newState

		# Custom to<State> function, or default exit function if none exists
		toFunc = getattr(self.states.get(oldState, None), f"to{newState}", self.getExitFunc(oldState))
		# Custom from<State> function, or default enter function if none exists
		fromFunc = getattr(self.states.get(newState, None), f"from{oldState}", self.getEnterFunc(newState))
		toFunc(*args)
		fromFunc(*args)
		# Always return true to prevent the base classes request method from calling it's class
		# private callEnterFunc and callExitFunc methods
		return True

	def getExitFunc(self, name: str) -> Callable:
		"""Returns the appropriate exit function when leaving a state, if it exists."""
		assert self.state == None and self.oldState == name

		# If there's no matching exitFoo() function, return defaultExit() instead.
		return getattr(self.states.get(name, None), "exit", self.defaultExit)