from random import randint
from typing import Literal, Optional, Sequence, Tuple, Union
from direct.showbase.ShowBase import ShowBase
import panda3d.core as p3d
from direct.gui.DirectGuiBase import DirectGuiWidget
import math
import io

from PIL import Image
from PIL.ImageDraw import Draw

import roboviz.Settings as Settings
from roboviz.Utils import p3d_to_rgb
from roboviz.controller.Controls import Controls as Ctrls, Game, OnlineVisualizer
from roboviz.model.Robot import Robot, RobotInstance
from roboviz.model.FileData import SwarmData as SData, PositionData as PData, TerrainData as TData
from roboviz.controller.FSM import FSM
from roboviz.view.States import *
from roboviz.view.Toaster import Toaster


class Scene(ShowBase):
	__instance = None
	__initialized = False
	DEFAULT_TERRAIN_SIZE = 50

	back_to_start = True
	"""
	Global property which determines whether States accessible from either Start or Menu return to
	Start or Menu when back is pressed. (inelegant, find better way)
	"""

	origin: p3d.NodePath
	"""A node at 0, 0, 0 that all other nodes are parented to"""
	toaster: Toaster
	"""The Scene's popup manager"""
	controls: Ctrls
	"""The Scene's control scheme manager"""
	state: FSM
	"""The Scene's state machine"""
	terrain: p3d.NodePath
	"""The NodePath that represents the Scene's terrain"""
	bot_pool: dict[str, Robot]
	"""The pool of robots available to the user to choose from"""
	instances: dict[str, RobotInstance]
	"""The robot instances currently being displayed in the scene"""
	layout: Literal["auto", "load"]
	"""The robot layout method being employed, either `"auto"` or `"load"` """

	def __new__(cls, *args: Any, **kwargs: Any):
		if Scene.__instance is None:
			Scene.__instance = super().__new__(cls)
		return Scene.__instance

	def __init__(self, robots: Optional[SData]=None, positions: Optional[PData]=None, terrain: Optional[TData]=None):
		if Scene.__initialized: return

		Scene.__initialized = True
		super().__init__()
		self.bot_pool = {}
		self.instances = {}
		self.layout = Settings.defaults.layout  # type: ignore
		self.origin = self.render.attach_new_node("origin")

		p3d.TextProperties.setDefaultFont(self.loader.loadFont(Settings.paths.font))
		self.overlay = DirectGuiWidget(frameSize=(-10, 10, -10, 10), frameColor=Settings.theme.bg[:3]+(0.6,))

		self.terrain = p3d.NodePath(p3d.Loader.get_global_ptr().load_sync(f"{Settings.paths.terrain}"))
		self.terrain.setTwoSided(True)
		self.terrain.setTransparency(p3d.TransparencyAttrib.M_alpha)
		self.terrain.setTexture(Scene.__create_terrain_texture(64, 4))
		self.terrain.setPos(0, 0, 0)
		self.terrain.reparentTo(self.origin)

		self.toaster = Toaster()

		self.disableMouse()		# disable default camera controls
		self.center_mouse()
		self.light_scene(80)  # Lights dont effect UI elements, so don't need to be removed and re-added when scene changes
		self.set_background_color(Settings.theme.bg)

		self.state = FSM("Application", [Start(), Controls(), Menu(), Visualizer(), Robots(), Positions()])
		# initializing state to be start menu on program start
		self.state.request("Start")

		self.load_demo(robots, positions, terrain)

		self.controls = Ctrls([Game(), OnlineVisualizer()], Settings.defaults.controls)

		self.run()

	def add_robot(self, bot: Robot) -> bool:
		"""Add a robot to library of available robots, return true if the robot is not a duplicate"""
		if not bot.is_valid:
			self.toaster.warn(f"Skipped robot ID: {bot.id} because it is invalid:\n{bot.error}")
			return False
		if bot.hash in self.bot_pool:
			self.toaster.warn(
				f"Skipped robot with ID {bot.id} because it is identical to robot number {self.bot_pool.get(bot.hash).number+1}", 5)
			return False
		bot.number = len(self.bot_pool)
		self.bot_pool[bot.hash] = bot
		return True

	def light_scene(self, elevation: int = 20) -> None:
		"""Add lights to the scene"""
		plight = self.render.attachNewNode(p3d.PointLight("pl1"))
		plight.setPos(0, 0, elevation)
		plight.look_at(0, 0, 0)
		self.render.setLight(plight)

		alight = p3d.AmbientLight("al1")
		alight.setColor(p3d.LVector4f(0.2, 0.2, 0.2, 1))
		self.render.setLight(self.render.attachNewNode(alight))

	def center_mouse(self) -> None:
		"""Center the mouse"""
		props = self.win.getProperties()  # type: ignore
		self.win.movePointer(0, props.getXSize() // 2, props.getYSize() // 2)  # type: ignore

	def load_demo(self, robots: Optional[SData]=None, positions: Optional[PData]=None, terrain: Optional[TData]=None) -> None:
		"""Load demo robots, or, if the user specified config files via the command line load those instead"""
		if robots == None or not robots.is_valid:
			if robots != None:
				self.toaster.warn(f"Invalid robot file: {robots.error}\n Loading default demo robots instead.")
			robots = SwarmData(Settings.paths.demo)
			if not robots.is_valid:
				self.toaster.warn(f"Invalid demo file: {robots.error}\nGo to CONFIGURE BOTS to load your own robots")
				self.scale_terrain(Scene.DEFAULT_TERRAIN_SIZE, Scene.DEFAULT_TERRAIN_SIZE)
				return
		for robot in robots.swarm:
			self.add_robot(Robot(robot))
		bot_ids = list(self.bot_pool.keys())
		if positions is None or not positions.is_valid:
			if positions != None:
				self.toaster.warn(f"Invalid position file: {positions.error}\nRobots will be auto positioned instead")
			if terrain != None:
				self.toaster.info("Terrain dimensions ignored because you can't specify dimensions in auto layout mode. If you wish to specify terrain dimensions, load a positions file from the CONFIGURE BOTS menu.", 15)
			# For however many robots there are, add that number + n to the scene where n is the
			# distance to the closest perfect square >= to the number of robots in the file. This
			# is to ensure that all robots are visible in the scene, and there is a perfect grid.
			self.auto_position(bot_ids, math.ceil(len(bot_ids)**0.5)**2, random=False)
		else:
			bot_size = self.max_bot_size(bot_ids)
			if terrain is None:
				self.scale_terrain(*positions.min_dimensions(bot_size))
			elif terrain.is_valid:
				if positions.fit_on(terrain.width, terrain.length, bot_size):
					self.scale_terrain(terrain.width, terrain.length)
				else:
					dimensions = positions.min_dimensions(bot_size)
					self.toaster.warn(f"Positions did not fit within terrain dimensions, auto sized to\n{dimensions[0]} x {dimensions[1]} instead")
					self.scale_terrain(*dimensions)
			else:
				dimensions = positions.min_dimensions(bot_size)
				self.toaster.warn(f"Failed to load terrain dimensions data:\n{terrain.error}\nAuto sized to {dimensions[0]} x {dimensions[1]}")
				self.scale_terrain(*dimensions)

			self.manual_position(bot_ids, positions.positions)

	def render_robot(self, id: str, pos: Tuple[float, float, float], rotation: int = 0, center: bool = False) -> RobotInstance:
		"""Render the given robot at the given position"""
		bot = self.bot_pool.get(id)
		assert bot is not None, f"Robot {id} not found"
		instance = bot.get_instance(str(len(self.instances)+1), (pos[0], pos[1], Settings.model.z_offset), rotation, center)
		instance.reparent_to(self.origin)
		self.instances[instance.id] = instance
		return instance

	def scale_terrain(self, width: float, length: Optional[float] = None) -> None:
		"""Scale the terrain to the given dimensions"""
		# setting terrain sz to one because if it is already the correct size, it will not scale
		if length == None: length = width
		length /= 2
		width /= 2
		self.terrain.setScale(width, length, 1)
		self.terrain.setTexScale(p3d.TextureStage.getDefault(), width, length)

	def auto_position(self, selected: list[str], number: int, random: bool = True) -> None:
		"""Auto position the selected robots in a grid"""
		n = number**0.5
		square = math.ceil(n)
		size = self.max_bot_size(selected)+Settings.robots.padding
		offset = (size*(square-1))/2
		self.scale_terrain(size*square)
		i = 0
		for row in range(square):
			for col in range(square):
				if i == number: break
				bot = selected[randint(0, len(selected)-1)] if random else selected[i%len(selected)]
				pos = (col*size - offset, row*size - offset, 0)
				self.render_robot(bot, pos, randint(0, 3)*90, center=True)
				i += 1

	def manual_position(self, selected: list[str], positions: list[tuple[float, float, float]]) -> bool:
		"""Position the selected robots at the given positions"""
		q = p3d.CollisionHandlerQueue()
		traverser = p3d.CollisionTraverser()
		for pos in positions:
			instance = self.render_robot(selected[randint(0, len(selected)-1)], pos)
			traverser.addCollider(instance.collider, q)
		traverser.traverse(self.render)
		traverser.clear_colliders()
		if q.get_num_entries() > 0:
			self.process_collision(q.entries, 0, None, selected, len(positions))
			return False
		return True

	def process_collision(self, entries: Sequence[p3d.CollisionEntry], index: int, remove: Optional[str], selected: list[str], total: int) -> None:
		"""Recursive function to handle collisions one at a time"""
		if remove != None:
			self.instances[remove].model.remove_node()
			del self.instances[remove]
		if index == len(entries):
			return
		from_bot = self.instances.get(entries[index].from_node.name)
		into_bot = self.instances.get(entries[index].into_node.name)
		if from_bot == None or into_bot == None:
			return self.process_collision(entries, index+1, None, selected, total)
		self.toaster.error(
			f"COLLISION:\n{from_bot} at {from_bot.pos} overlaps\n{into_bot} at {into_bot.pos}.",
			[f"Remove #{from_bot.id}", f"Remove #{into_bot.id}", "Auto Position"],
			[self.process_collision, self.process_collision, self.auto_instead],
			[
				(entries, index+1, from_bot.id, selected, total),
				(entries, index+1, into_bot.id, selected, total),
				(selected, total)
			]
		)

	def auto_instead(self, selected: list[str], number: int) -> None:
		"""
		Function that runs when the user chooses to auto position instead when there is a collision
		"""
		self.clear_instances()
		self.auto_position(selected, number)

	def clear_instances(self):
		"""Clear all instances from the scene"""
		for i in self.instances.values():
			i.model.remove_node()
		self.instances = {}

	def max_bot_size(self, bot_ids: Union[list[str], set[str]]) -> float:
		"""Return the maximum dimension of the largest bot in a list of bot ids"""
		result = 0
		for id in bot_ids:
			bot = self.bot_pool.get(id)
			if bot != None:
				result = max(result, bot.width, bot.length)
		return result

	def orbit(self, task: Task.Task) -> int:
		"""Orbit the camera around the scene"""
		self.camera.set_x(self.camera, self.terrain.get_sx()/100) # type: ignore
		self.camera.look_at(0, 0, 0)
		return Task.cont

	def angle_camera(self):
		"""Angle the camera in a cinematic manner"""
		self.camera.set_pos(self.terrain.get_sx()*2.5, self.terrain.get_sx()*2.5, self.terrain.get_sx()*2)
		self.camera.look_at(0, 0, 0)

	def toggle_bounding_boxes(self):
		"""Toggle the visibility of the bounding boxes of the rendered models"""
		for i in self.instances.values():
			if i.collider.is_hidden():
				i.collider.show()
			else:
				i.collider.hide()

	@staticmethod
	def __create_terrain_texture(size: int, line_width: int) -> p3d.Texture:
		"""
		Generate a `+` texture to be used for the terrain.

		`size` - The size of the image to apply to the texture, greater size == crisper texture.

		`line_width` - The width of the grid lines. The greater the size the thinner these will appear.
		"""
		img = Image.new("RGBA", (size, size), (0, 0, 0, 0))
		canvas = Draw(img)
		canvas.line((size//2, 0, size//2, size), fill=p3d_to_rgb(Settings.theme.gui.bg_disabled), width=line_width)
		canvas.line((0, size//2, size, size//2), fill=p3d_to_rgb(Settings.theme.gui.bg_disabled), width=line_width)

		byte_arr = io.BytesIO()
		img.save(byte_arr, format='PNG')

		p = p3d.PNMImage()
		p.read(p3d.StringStream(byte_arr.getvalue()))
		tex = p3d.Texture()
		tex.load(p)
		return tex