import json
from typing import Any, Tuple
from abc import ABC
from os.path import exists, basename, normpath
import roboviz.Settings as Settings

__all__ = ["PositionData", "SwarmData", "TerrainData"]

class FileData(ABC):
	path: str
	is_valid: bool
	_error: str
	file_name: str

	def __init__(self, path: str):
		self.is_valid = True
		self.path = normpath(path)
		self.file_name = basename(path)
		self.parse()

	def parse(self) -> list[str]:
		if not exists(self.path):
			self.error = f"File {self.path} does not exist"
			return []
		try:
			with open(self.path, "r") as f:
				return [line for line in f if not line.lstrip().startswith("#") and not line.lstrip().startswith("//")]
		except PermissionError:
			self.error = "No file selected"
			return []

	@property
	def error(self): return self._error

	@error.setter
	def error(self, val: str):
		self.is_valid = False
		self._error = val


class TerrainData(FileData):
	data: dict[str, int]
	width: int
	length: int

	def __init__(self, path: str):
		super().__init__(path)

	def parse(self) -> None:
		lines = super().parse()
		if not self.is_valid: return
		try:
			self.length = int(lines[0])
			self.width = int(lines[1])
			if self.width <= 0 or self.length <= 0:
				self.error = "Terrain dimensions cannot be 0 or negative"
		except Exception as e:
			if type(e) == ValueError:
				self.error = "Invalid file format, terrain dimensions should in the form:\n50\n50"
			elif type(e) == IndexError:
				self.error = "Missing length or width"
			else:
				self.error = str(e)


class PositionData(FileData):
	total: int
	positions: list[tuple[float, float, float]]
	n: float
	e: float
	s: float
	w: float

	def __init__(self, path: str):
		self.positions = []
		self.n = 0
		self.e = 0
		self.s = 0
		self.w = 0
		super().__init__(path)

	def parse(self) -> None:
		lines = super().parse()
		self.total = len(lines)
		if not self.is_valid: return
		for line in lines[:Settings.robots.max]:
			try:
				coords = line.split()
				self.positions.append((float(coords[0]), float(coords[1]), float(coords[2])))
				# Establish the furthest point in each direction
				if self.positions[-1][0] > self.e: self.e = self.positions[-1][0]
				if self.positions[-1][0] < self.w: self.w = self.positions[-1][0]
				if self.positions[-1][1] > self.n: self.n = self.positions[-1][1]
				if self.positions[-1][1] < self.s: self.s = self.positions[-1][1]
			except Exception as e:
				if type(e) == ValueError:
					self.error = "Invalid positions in file"
				elif type(e) == IndexError:
					self.error = "Positions are not in the correct format. They should be in the form x y z, with each new position on a new line."
				else:
					self.error = str(e)
				break

	def fit_on(self, width: float, length: float, margin: float = 0) -> bool:
		"""
		Return true if all the coordinates fit within a terrain with the given width and length
		a given margin from the edge
		"""
		if self.e + margin > width/2: return False
		if abs(self.w) + margin > width/2: return False
		if self.n + margin > length/2: return False
		if abs(self.s) + margin > length/2: return False
		return True

	def min_dimensions(self, margin: float = 0) -> Tuple[int, int]:
		"""
		Return the minimum terrain dimensions that would fit all of the positions in it with a
		margin of `margin` from each edge.
		"""
		return (round(max(self.e, abs(self.w))+margin)*2, round(max(self.n, abs(self.s))+margin)*2)


class SwarmData(FileData):
	swarm: list[dict[str, Any]]

	def __init__(self, path: str):
		self.swarm = []
		super().__init__(path)

	def parse(self) -> None:
		lines = super().parse()
		if not self.is_valid: return
		try:
			swarm = json.loads("\n".join(lines))
			if "swarm" in swarm:
				self.swarm = swarm["swarm"]
			# if the file contains a single robot instead of a swarm
			elif isinstance(swarm, dict):
				self.swarm = [swarm]
		except Exception as e:
			if type(e) == json.JSONDecodeError:
				self.error = "File is in invalid JSON format"
			else:
				self.error = str(e)