from __future__ import annotations
from abc import ABC	# required in order to use the classes own type inside the class
from typing import Literal, Optional, Tuple
from panda3d.core import NodePath, Loader
from random import randint

import roboviz.controller.Scene as S
from PIL import Image
from PIL.ImageDraw import Draw
from PIL.Image import Image as Img

import roboviz.Settings as Settings
from roboviz.Utils import p3d_to_rgb, rotate_img

class Part(ABC):
	SRC_ROTATIONS = {0: 180, 1: 0, 2: 90, 3: 270}
	"""The degrees to rotate a part by for a given source slot"""
	DEST_ROTATIONS = {0: 0, 1: 180, 2: 90, 3: 270}
	"""The degrees to rotate a part by for a given destination slot"""
	DEFAULT_COLOR = (0.2, 0.2, 0.2)
	colors: dict[str, Tuple[float, float, float]] = {}
	img = Image.new("RGB", (16, 16), p3d_to_rgb(Settings.theme.gui.bg))

	id: str
	orientation: int
	offset: float
	has_slots: bool
	absolute_rotation: int
	dest_slot: int
	__hash: Optional[str]
	size: float
	model: NodePath
	parent: Optional[Part]
	children: list[Optional[Part]]

	def __init__(self, id: str, model_type: str, orientation: int):
		self.id = id
		self.orientation = orientation
		self.offset = 41
		self.has_slots = True
		self.absolute_rotation = 0
		self.dest_slot = 0
		self.__hash = None
		self.model = NodePath(Loader.get_global_ptr().load_sync(f"{Settings.model.path}/{model_type}.{Settings.model.extension}"))

		self.model.set_color(*self.color)

		self.children = [None, None, None, None]
		self.parent = None

	def reparent_to(self, parent: Part, src_slot: Literal[0, 1, 2, 3], dest_slot: Literal[0, 1, 2, 3]):
		parent.children[src_slot] = self
		self.parent = parent
		self.position(src_slot, dest_slot, parent)
		self.absolute_rotation = parent.absolute_rotation + Part.SRC_ROTATIONS[src_slot] + Part.DEST_ROTATIONS[dest_slot]
		self.absolute_rotation %= 360
		self.dest_slot = dest_slot

		self.model.reparent_to(parent.model)
		self.model.set_r(self.orientation*90)
		self.__hash = None

	def position(self, src_slot: int, dest_slot: int, parent: Part):
		# note: HPR stands for Heading, Pitch, Roll. It should be Yaw, Pitch, Roll, but y is already used for the y axis
		if src_slot == 0:
			self.model.setPos(0, self.get_offset(parent), 0)
		elif src_slot == 1:
			self.model.setPos(0, -self.get_offset(parent), 0)
			self.model.set_h(self.model, 180)
		elif src_slot == 2:
			self.model.setPos(-self.get_offset(parent), 0, 0)
			self.model.set_h(self.model, 90)
		elif src_slot == 3:
			self.model.setPos(self.get_offset(parent), 0, 0)
			self.model.set_h(self.model, 270)

		if dest_slot == 0:
			self.model.set_h(self.model, 180)
		elif dest_slot == 3:
			self.model.set_h(self.model, 270)
		elif dest_slot == 2:
			self.model.set_h(self.model, 90)

	def __str__(self):
		return f"{self.id}: {type(self).__name__}[{','.join([type(child).__name__ if child else 'None' for child in self.children])}]"

	def draw_to(self, canvas: Img, coords: Tuple[int, int]):
		canvas.paste(rotate_img(self.img, self.absolute_rotation), coords)
		for child in self.children:
			if child: child.draw_to(canvas, Part.translate(child.dest_slot, child.absolute_rotation, coords))

	def get_offset(self, parent: Part) -> float:
		"""
		Return the correct offset value for this part and the given parent part. If one of them is
		slotted and the other isn't reduce the offset to make the parts fit together.
		"""
		offset = self.offset/2 + parent.offset/2
		if parent.has_slots ^ self.has_slots: offset -= 1.5		# ^ is xor
		return offset

	@property
	def color(self) -> Tuple[float, float, float]:
		if type(self).__name__ not in Part.colors:
			if len(Settings.theme.part.colors) > 0:
				Part.colors[type(self).__name__] = Settings.theme.part.colors.pop(randint(0, len(Settings.theme.part.colors)-1))
			else:
				Part.colors[type(self).__name__] = Part.DEFAULT_COLOR
		return Part.colors[type(self).__name__]

	@property
	def hash(self):
		if not self.__hash:
			self.__hash = f"{type(self).__name__}{self.orientation}[{','.join([child.hash if child else 'None' for child in self.children])}]"
		return self.__hash


	@staticmethod
	def translate(dest_slot: int, rotation: int, coords: Tuple[int, int]) -> Tuple[int, int]:
		if dest_slot == 1: rotation += 180
		if dest_slot == 2: rotation += 270
		if dest_slot == 3: rotation += 90
		rotation %= 360
		x, y = 0, 0
		if rotation == 0: y += 16
		if rotation == 90: x -= 16
		if rotation == 180: y -= 16
		if rotation == 270: x += 16
		return (coords[0]+x, coords[1]+y)


class CoreComponent(Part):
	__img: Optional[Img] = None

	def __init__(self, id: str, orientation: int):
		super().__init__(id, type(self).__name__, orientation)

	@property
	def img(self) -> Img:
		"""Lazy load image, and create it for class not instance to save memory"""
		if not CoreComponent.__img:
			CoreComponent.__img = Part.img.copy()
			Draw(CoreComponent.__img).rectangle((0, 0, 15, 15), fill=p3d_to_rgb(Settings.theme.gui.fg), outline=p3d_to_rgb(self.color), width=4)
		return CoreComponent.__img

class FixedBrick(Part):
	__img: Optional[Img] = None

	def __init__(self, id: str, orientation: int):
		super().__init__(id, type(self).__name__, orientation)

	def reparent_to(self, parent: Part, src_slot: Literal[0, 1, 2, 3], dest_slot: Literal[0, 1, 2, 3]) -> None:
		super().reparent_to(parent, src_slot, dest_slot)
		# Using Scene singleton to access global render
		self.model.set_r(-parent.model.get_r(S.Scene().render))

	@property
	def img(self):
		"""Lazy load image, and create it for class not instance to save memory"""
		if not FixedBrick.__img:
			FixedBrick.__img = Part.img.copy()
			Draw(FixedBrick.__img).rectangle((0, 0, 15, 15), outline=p3d_to_rgb(self.color), width=4)
		return FixedBrick.__img

class ActiveHinge(Part):
	__img: Optional[Img] = None

	def __init__(self, id: str, orientation: int):
		super().__init__(id, type(self).__name__, orientation)
		self.offset = 38.75
		self.has_slots = False

	@property
	def img(self) -> Img:
		"""Lazy loaded 2d image representation of part"""
		if not ActiveHinge.__img:
			ActiveHinge.__img = Part.img.copy()
			color = p3d_to_rgb(self.color)
			canvas = Draw(ActiveHinge.__img)
			canvas.rectangle((0, 0, 15, 1), fill=color)
			canvas.rectangle((0, 14, 15, 15), fill=color)
			canvas.line((7, 0, 7, 15), fill=color, width=4)
			canvas.line((7, 8, 7, 15), fill=color, width=10)
		return ActiveHinge.__img

class PassiveHinge(Part):
	__img: Optional[Img] = None

	def __init__(self, id: str, orientation: int):
		super().__init__(id, type(self).__name__, orientation)
		self.offset = 40
		self.has_slots = False

	@property
	def img(self) -> Img:
		"""Lazy loaded 2d image representation of part"""
		if not PassiveHinge.__img:
			PassiveHinge.__img = Part.img.copy()
			color = p3d_to_rgb(self.color)
			canvas = Draw(PassiveHinge.__img)
			canvas.rectangle((0, 0, 15, 1), fill=color)
			canvas.rectangle((0, 14, 15, 15), fill=color)
			canvas.line((7, 0, 7, 15), fill=color, width=4)
		return PassiveHinge.__img