from __future__ import annotations
from typing import Optional, Tuple
from PIL import Image
from PIL.Image import Image as Img
from panda3d.core import Texture, PNMImage, StringStream, LPoint3f, CollisionBox, CollisionNode
import io

from roboviz.model.Part import *
from roboviz.Utils import p3d_to_rgb
import roboviz.Settings as Settings

class Robot:
	SCALE = 0.024390233
	"""The scale at which a CoreComponent is 1 x 1"""
	PART_TYPES = {"CoreComponent": CoreComponent, "FixedBrick": FixedBrick, "ActiveHinge": ActiveHinge, "PassiveHinge": PassiveHinge}

	number: Optional[int]
	"""Used externally by the Scene class to keep track of how many robot types there are"""
	id: str
	root: Optional[Part]
	parts: dict[str, Part]
	is_valid: bool
	error: str
	width: float
	length: float
	__hash: Optional[str]
	__img: Optional[Img]
	__tex: Optional[Texture]

	def __init__(self, data: dict):
		self.number = None
		self.root = None
		self.__img = None
		self.__hash = None
		self.__tex = None
		self.parts = {}
		self.is_valid = True
		# todo ensure that dict has the right keys

		self.id = str(data['id'])
		for part in data["body"]["part"]:
			self.add_part(part)
			if not self.is_valid: return
			if part["root"]:
				self.root = self.parts[part["id"]]
		for con in data["body"]["connection"]:
			self.add_connection(con)
			if not self.is_valid: return
		self.model.set_scale(Robot.SCALE)

		b: Tuple[LPoint3f, LPoint3f] = self.model.get_tight_bounds()
		self.width = abs(b[0].get_x()-b[1].get_x())
		self.length = abs(b[0].get_y()-b[1].get_y())

	def add_part(self, part: dict) -> None:
		if part["type"] not in Robot.PART_TYPES:
			self.error = f"Part {part['type']} has not been implemented"
			self.is_valid = False
			return
		self.parts[part["id"]] = Robot.PART_TYPES[part["type"]](part["id"], part["orientation"])

		self.__hash = None
		self.__img = None
		self.__tex = None

	def add_connection(self, con: dict) -> None:
		dest = self.parts.get(con["dest"], None)
		src = self.parts.get(con["src"], None)
		if not dest:
			self.error = f"Part {con['dest']} does not exist"
			self.is_valid = False
			return
		if not src:
			self.error = f"Part {con['src']} does not exist"
			self.is_valid = False
			return
		dest.reparent_to(src, con["srcSlot"], con["destSlot"])
		self.__hash = None
		self.__img = None
		self.__tex = None

	@property
	def model(self) -> NodePath: return self.root.model

	@property
	def hash(self) -> str:
		"""Unique hash of the robot determined by it's structure, used to compare robots"""
		if not self.__hash: self.__hash = self.root.hash
		return self.__hash

	def __eq__(self, __o: object) -> bool:
		return isinstance(__o, Robot) and self.hash == __o.hash

	@property
	def texture_2d(self) -> Texture:
		"""Panda3D texture created from a 2D image of the robot"""
		if not self.__img or not self.__tex:
			self.__img = Image.new('RGB', (256, 256), p3d_to_rgb(Settings.theme.gui.bg))
			self.root.draw_to(self.__img, (119, 119))

			byte_arr = io.BytesIO()
			self.__img.save(byte_arr, format='PNG')

			p = PNMImage()
			p.read(StringStream(byte_arr.getvalue()))
			self.__tex = Texture()
			self.__tex.load(p)
		return self.__tex

	def get_instance(self, id: str, pos: Tuple[float, float, float], rotation: int = 0, center: bool = False) -> RobotInstance:
		return RobotInstance(id, self, pos, rotation, center)

class RobotInstance:
	"""A lighter weight version of Robot containing only the information needed to display it"""
	id: str
	type: str
	pos: Tuple[float, float]
	width: float
	length: float
	model: NodePath
	def __init__(self, id: str, robot: Robot, pos: Tuple[float, float, float], rotation: int = 0, center: bool = False):
		self.id = id
		self.type = robot.id
		self.pos = pos[:2]
		self.model = robot.model.__copy__()
		self.model.set_h(rotation)
		bounds: Tuple[LPoint3f, LPoint3f] = self.model.get_tight_bounds()
		self.width = abs(bounds[0].get_x()-bounds[1].get_x())
		self.length = abs(bounds[0].get_y()-bounds[1].get_y())
		if center:
			self.model.set_pos(pos[0]-(self.width/2+bounds[0].get_x()), pos[1]-(self.length/2+bounds[0].get_y()), pos[2])
		else:
			self.model.set_pos(*pos)
		self.collider = self.model.attach_new_node(CollisionNode(self.id))
		self.collider.node().add_solid(CollisionBox(*bounds))
		self.collider.set_h(-rotation)
		self.collider.set_scale(1/Robot.SCALE)

	def reparent_to(self, parent: NodePath) -> None:
		"""
		Removes the referenced node of the NodePath from its current parent and attaches it to the referenced node of the indicated NodePath.

		If the referenced node is already a child of the indicated NodePath (via some other instance), this operation fails and leaves the NodePath detached.
		"""
		self.model.reparent_to(parent)

	def __str__(self) -> str:
		return f"Bot #{self.id}, Type ID: {self.type}"