from __future__ import annotations
from typing import Any, Optional, Callable, Tuple
from direct.gui.DirectGui import DirectButton, DirectSlider, OnscreenImage, OnscreenText, DirectScrolledFrame, DirectEntry
import direct.gui.DirectGuiGlobals as DGG
from abc import ABC, abstractmethod
from panda3d.core import TransparencyAttrib, Vec4, PNMImage, StringStream, Texture, PGMouseWatcherParameter
from PIL import Image, ImageDraw
import io
import roboviz.controller.Scene as S

from roboviz.Utils import p3d_to_rgb
import roboviz.Settings as Settings


class Widget(ABC):
	"""
	# Widget
	All classes that implement the Widget interface have a destroy method that removes the widget
	from the scene, and access to the global scene object.
	"""
	@abstractmethod
	def destroy(self) -> None: pass

	@property
	def scene(self) -> S.Scene: return S.Scene()

class Button(DirectButton, Widget):
	SCALE_MULTIPLIER = 5
	def __init__(self, text: str, x: float =0, y: float =0, scale: float =1, width: float = 2.5, **kwargs: Any):
		super().__init__(
			text=text.upper(),
			relief=DGG.FLAT,
			text_fg=Settings.theme.gui.fg,
			frameColor=Settings.theme.gui.bg,
			scale=scale/Button.SCALE_MULTIPLIER,
			text_scale=0.5,
			frameSize=(-width, width, -0.25, 0.6),
			pos=(x, 0, y),
			**kwargs
		)
		self.initialiseoptions(Button)

	def enable(self) -> None:
		self["state"] = DGG.NORMAL
		self["frameColor"] = Settings.theme.gui.bg
		self["text_fg"] = Settings.theme.gui.fg

	def disable(self) -> None:
		self["state"] = DGG.DISABLED
		self["frameColor"] = Settings.theme.gui.bg_disabled
		self["text_fg"] = Settings.theme.gui.fg_disabled


class Select(Button):
	"""
	A class representing an options menu, which takes a list of tuples as options (name, value)
	"""
	# todo Scale
	OPEN_SYMBOL = "-"
	CLOSED_SYMBOL = "+"
	OFFSET = -0.8
	WIDTH = 3
	__option_objects: list[Button]
	is_open: bool
	command: Callable[[str], Any]
	options: dict[str, str]
	def __init__(self,
			x: float,
			y: float,
			options: list[Tuple[str, str]],
			command: Callable[[str], Any],
			selected: Optional[str] = None,
			scale: float = 1,
			**kwargs: Any):

		selected = selected or options[0][1]
		self.options = {val: name for name, val in options}
		super().__init__(Select.format(self.options[selected], Select.CLOSED_SYMBOL), x, y, scale, width=Select.WIDTH, command=self.on_click, **kwargs)
		self.command = command
		self.selected = selected
		self.is_open = False
		self.__option_objects = []
		self.initialiseoptions(Select)

	def on_click(self) -> None:
		if self.is_open:
			self.close()
		else:
			self.setText(self.format(self.selected, Select.OPEN_SYMBOL))
			offset = Select.OFFSET
			for option in self.options.keys():
				btn = Button(
					self.format(self.options[option]),
					0, offset,
					width = Select.WIDTH,
					command = self.on_select,
					extraArgs = [option],
					scale=Button.SCALE_MULTIPLIER,
				)
				offset += Select.OFFSET
				if option == self.selected:
					btn.set_color(Vec4(*Settings.theme.gui.bg_secondary))
				self.__option_objects.append(btn)
				btn.reparent_to(self)
			self.is_open = True

	def on_select(self, selected: str) -> None:
		self.selected = selected
		self.command(selected)
		self.close()

	def close(self) -> None:
		self.setText(self.format(self.options[self.selected], Select.CLOSED_SYMBOL))
		self.is_open = False
		for btn in self.__option_objects:
			btn.destroy()

	def add_option(self, option: Tuple[str, Any]) -> None:
		self.options[option[0]] = option[1]

	@staticmethod
	def format(text: str, symbol: str = "") -> str:
		# todo truncate and add dots once I know what max length is
		return f"{text.upper():<20}{symbol}"


class Slider(DirectSlider, Widget):
	MAX = 100
	MIN = 0

	fn: Callable[[float], None]
	base_text: str
	def __init__(self, text: str,
			x: float, y: float,
			command: Callable[[float], None],
			default: int =50,
			minimum: int = 0,
			maximum: int = 0,
			**kwargs: Any):
		self.fn = command
		self.base_text = text.upper()
		self.total = (maximum or Slider.MAX) - (minimum or Slider.MIN)
		super().__init__(
			text=f"{self.base_text} ({default}%)",
			text_pos=(0, 0.1),
			text_scale=0.15,
			text_fg=Settings.theme.fg,
			range=(minimum or Slider.MIN, maximum or Slider.MAX),
			value=default,
			pageSize=5,
			command=self.update,

			frameColor=Settings.theme.gui.bg,
			scale=1,
			thumb_relief = DGG.FLAT,
			thumb_frameColor = Settings.theme.gui.fg,
			pos=(x, 0, y),
			**kwargs
		)
		self.initialiseoptions(Slider)

	def update(self) -> None:
		self["text"] = f"{self.base_text} ({round(self['value'])}%)"
		self.fn(int(self["value"])/self.total)

class Logo(Widget):
	left_eyes: list[OnscreenImage]
	right_eyes: list[OnscreenImage]
	left_eye: OnscreenImage
	right_eye: OnscreenImage
	left_index: int
	right_index: int
	left_rotation: float
	right_rotation: float
	def __init__(self, x: float, y: float):
		super().__init__()
		self.left_index, self.right_index = 1, 1
		self.left_rotation, self.right_rotation = 0, 0
		self.left_eyes = []
		self.right_eyes = []
		self.logo = OnscreenImage(image=f"{Settings.paths.logo}/logo.png", scale=(0.8, 0, 0.5), pos=(x, 0, y))
		self.logo.setTransparency(TransparencyAttrib.M_alpha)
		for i in range(0, 2):
			left_eye = OnscreenImage(image=f"{Settings.paths.logo}/eye-{i}.png", scale=(0.153, 0, 0.153), pos=(x+0.26, 0, y-0.025))
			left_eye.setTransparency(TransparencyAttrib.M_alpha)
			left_eye.hide()
			self.left_eyes.append(left_eye)

			right_eye = OnscreenImage(image=f"{Settings.paths.logo}/eye-{i}.png", scale=(0.153, 0, 0.153), pos=(x+0.645, 0, y-0.025))
			right_eye.setTransparency(TransparencyAttrib.M_alpha)
			right_eye.hide()
			self.right_eyes.append(right_eye)

		self.left_eye = self.left_eyes[self.left_index]
		self.left_eye.show()
		self.right_eye = self.right_eyes[self.right_index]
		self.right_eye.show()

	def destroy(self) -> None:
		self.logo.destroy()
		for left, right in zip(self.left_eyes, self.right_eyes):
			left.destroy()
			right.destroy()

	def look(self, x: float, y: float) -> None:
		if x < 0:
			if y > 0.525:
				self.set_left(0, 90)
				self.set_right(0, 90)
			elif y > 0.22:
				self.set_left(1, 90)
				self.set_right(1, 90)
			else:
				self.set_left(0, 0)
				self.set_right(0, 0)
		elif x > 0.5:
			if y > 0.525:
				self.set_left(0, 180)
				self.set_right(0, 180)
			elif y > 0.22:
				self.set_left(1, -90)
				self.set_right(1, -90)
			else:
				self.set_left(0, -90)
				self.set_right(0, -90)
		elif x > 0:
			if y > 0.625:
				self.set_left(1, 180)
				self.set_right(1, 180)
			if y > 0.525:
				self.set_left(0, 180)
				self.set_right(0, 90)
			elif y > 0.22:
				self.set_left(1, -90)
				self.set_right(1, 90)
			elif y > 0.1:
				self.set_left(0, -90)
				self.set_right(0, 0)
			else:
				self.set_left(1, 0)
				self.set_right(1, 0)

	def set_left(self, index: int, rotation: int) -> None:
		if self.left_index == index and self.left_rotation == rotation: return
		self.left_rotation = rotation
		self.left_eyes[index].set_r(rotation)
		self.left_eyes[index].show()

		if self.left_index != index:
			self.left_index = index
			self.left_eye.hide()
			self.left_eye.set_r(0)
			self.left_eye = self.left_eyes[index]

	def set_right(self, index: int, rotation: int) -> None:
		if self.right_index == index and self.right_rotation == rotation: return
		self.right_rotation = rotation
		self.right_eyes[index].set_r(rotation)
		self.right_eyes[index].show()

		if self.right_index != index:
			self.right_index = index
			self.right_eye.hide()
			self.right_eye.set_r(0)
			self.right_eye = self.right_eyes[index]


class Text(OnscreenText, Widget):
	def __init__(self, text: str, x: float, y: float, scale: float, **kwargs: Any):
		super().__init__(
			text=text,
			pos=(x, y),
			fg=Settings.theme.fg,
			scale=scale/10,
			**kwargs
		)


class Options(DirectScrolledFrame, Widget):
	X_START = -0.8
	Y_START = 0.61
	X_STEP = 0.26
	Y_STEP = 0.24
	MAX_COLS = 7
	row: int
	col: int
	shift_down: bool
	options: dict[str, Options.Option]
	options_list: list[Options.Option]
	last_clicked: int
	selected: set[str]
	add_btn: DirectButton
	command: Optional[Callable[[int], Any]]
	def __init__(self, x: float, y: float, add_fn: Optional[Callable] =None, command: Optional[Callable] =None, **kwargs: Any):
		super().__init__(
			frameSize=(-1, 1, -0.5, 0.5),
			canvasSize=(-1, 0.9, -0.21, 0.8),
			color=Settings.theme.gui.bg_disabled[:3]+(0.2,),
			pos=(x, 0, y),
			verticalScroll_frameSize=(-0.03, 0.03, -0.3, 0.3),
			**kwargs
		)
		self.row, self.col = 0, 0
		self.shift_down = False
		self.options = {}
		self.options_list = []
		self.last_clicked = 0
		self.selected = set()
		self.command = command
		self.initialiseoptions(Options)
		self.set_transparency(TransparencyAttrib.M_alpha)

		im = Image.new("RGB", (64, 64), p3d_to_rgb(Settings.theme.gui.bg))
		draw = ImageDraw.Draw(im)
		draw.line((31, 10, 31, 53), fill=p3d_to_rgb(Settings.theme.gui.fg), width=8)
		draw.line((10, 31, 53, 31), fill=p3d_to_rgb(Settings.theme.gui.fg), width=8)
		byte_arr = io.BytesIO()
		im.save(byte_arr, format="PNG")

		p = PNMImage()
		p.read(StringStream(byte_arr.getvalue()))
		tex = Texture()
		tex.load(p)

		self.add_btn = DirectButton(
			image=tex,
			pos=(self.next_x, 0, self.next_y),
			relief=DGG.FLAT,
			text_fg=Settings.theme.gui.fg,
			scale=0.1,
			# text_scale=2,
			command=add_fn,
			pressEffect=0,
			parent=self.getCanvas(),
			**kwargs
		)
		self.add_btn.set_transparency(TransparencyAttrib.M_none)
		self.add_btn.set_color_off()

	def add_option(self, hash: str, texture: Texture) -> None:
		o = Options.Option(self, len(self.options), texture, hash, self.next_x, self.next_y)
		o.reparent_to(self.getCanvas())
		self.options[o.id] = o
		self.options_list.append(o)
		self.col += 1
		if self.col == Options.MAX_COLS:
			self.col = 0
			self.row += 1
			self["canvasSize"] = (-1, 0.9, min(-0.21,-self.row*0.2), 0.8)
		self.add_btn.set_pos(self.next_x, 0, self.next_y)

	def toggle_all(self, state: bool) -> None:
		for option in self.options.values():
			option.toggle(state)
		self.last_clicked = 0

	def toggle(self, hash: str, state: Optional[bool] =None) -> None:
		if hash in self.options:
			self.options.get(hash).toggle(state)

	@property
	def next_x(self) -> float: return Options.X_START+self.col*Options.X_STEP

	@property
	def next_y(self) -> float: return Options.Y_START-self.row*Options.Y_STEP

	@property
	def total_selected(self) -> int: return len(self.selected)

	class Option(DirectButton, Widget):
		is_selected: bool
		instance: int
		__parent: Options
		id: str
		def __init__(self, parent: Options, instance: int, texture: Texture, id: str, x: float, y: float, **kwargs: Any):
			super().__init__(
				image=texture,
				pos=(x, 0, y),
				relief=DGG.FLAT,
				text_fg=Settings.theme.gui.fg,
				scale=0.1,
				text_scale=0.6,
				frameSize=(-1, 1, -0.8, 1),
				pressEffect=False,
				command=self.toggle,
				**kwargs
			)
			self.__parent = parent
			self.instance = instance
			self.id = id
			self.is_selected = False
			self.set_transparency(TransparencyAttrib.M_none)
			self.set_color_off()
			self.initialiseoptions(Options.Option)

		def toggle(self, state: Optional[bool]=None) -> None:
			prev = self.is_selected
			if self.__parent.shift_down:
				direction = 1 if self.instance > self.__parent.last_clicked else -1
				for i in range(self.__parent.last_clicked, self.instance, direction):
					self.__parent.options_list[i].toggle(True)
				self.is_selected = True
			else:
				self.is_selected = not self.is_selected if state == None else state
			if self.is_selected:
				self.set_scale(0.09)
				self.set_color(*Settings.theme.gui.bg_secondary)
			else:
				self.set_scale(0.1)
				self.set_color_off()
			self.__parent.last_clicked = self.instance

			if prev != self.is_selected:
				if self.is_selected:
					self.__parent.selected.add(self.id)
				else:
					self.__parent.selected.remove(self.id)
				if self.__parent.command: self.__parent.command(self.__parent.total_selected)

class Entry(DirectEntry, Widget):
	SCALE_MULTIPLIER = 10

	_text: str
	callback: Callable
	filter: set[int]

	def __init__(self,
			initial_text: str,
			filter: str,
			x: float, y: float,
			width: float, scale: float,
			callback: Callable,
			**kwargs: Any):
		self.callback = callback
		super().__init__(
			initialText=initial_text,
			pos=(x, 0, y),
			width=width,
			scale=scale/Entry.SCALE_MULTIPLIER,
			text_fg=Settings.theme.fg,
			frameColor=Settings.theme.gui.bg_disabled[:3]+(0.5,),
			focusInCommand=lambda: self.accept("delete", self._on_del),
			focusOutCommand=lambda: self.ignore("delete"),
			command=lambda: self.ignore("delete"),
			**kwargs
		)
		self._text = initial_text
		self.filter = set([ord(i) for i in filter])
		self.filter.add(8) # Backspace
		self.accept(self.guiItem.getKeystrokeEvent(), self.on_key)
		self.initialiseoptions(Entry)

	@property
	def text(self) -> str: return self._text

	@text.setter
	def text(self, val: str) -> None:
		self._text = val
		self.set(val)
		self.callback(val)

	def on_key(self, key: PGMouseWatcherParameter) -> None:
		"""
		Called on each keystroke event. If the key is in the filter, it is added to the text. If not
		the text is reset to what it was before the keystroke, and the cursor is moved back
		"""
		if self.valid(key):
			self.text = self.get()
		else:
			self.set(self.text)
			if self.getCursorPosition() != len(self.get()):
				self.setCursorPosition(self.getCursorPosition()-1)

	def _on_del(self) -> None:
		"""
		For internal use only. The delete key does not seem to send a keystroke event, so it needs to
		be handled manually
		"""
		self.text = self.get()

	def valid(self, key: PGMouseWatcherParameter) -> bool:
		return len(self.text) == len(self.get()) or key.get_keycode() in self.filter


class NumEntry(Entry):
	max_num: int

	def __init__(self,
			initial_val: int,
			max_num: int,
			x: float, y: float,
			width: float, scale: float,
			callback: Callable,
			**kwargs: Any):
		"""Settings `max_num` to 0 will allow any number"""
		super().__init__(str(initial_val), "0123456789", x, y, width, scale, callback, **kwargs)
		self.initialiseoptions(NumEntry)
		self.max_num = max_num

	def valid(self, key: PGMouseWatcherParameter) -> bool:
		if not super().valid(key): return False
		value = int(self.get()) if self.get() else 0
		if value > self.max_num:
			if self.max_num > 0:
				self.scene.toaster.warn(f"{value} exceeds maximum of {self.max_num}, value set to {self.max_num} instead.")
				self.text = str(self.max_num)
		return True

	@property
	def text(self) -> str: return self._text

	@text.setter
	def text(self, val: str) -> None:
		self._text = val
		self.set(val)
		self.callback(int(val) if val else 0)