import math
from typing import Any, Optional, Tuple
from direct.task import Task
from direct.gui.DirectFrame import DirectFrame
from direct.gui.DirectGuiBase import DirectGuiWidget

import roboviz.Settings as Settings
from roboviz.model.Robot import Robot
from roboviz.model.FileData import TerrainData, PositionData, SwarmData
from roboviz.controller.FSM import State
from roboviz.view.Components import Button, NumEntry, Options, Slider, Logo, Text, Select

from tkinter import filedialog as fd


class Start(State):
	def enter(self) -> None:
		self.logo = Logo(0, 0.425)
		self.add_widget(self.logo)
		self.add_widget(Button("View", 0, -0.25, command=self.request, extraArgs=["Visualizer"]))
		self.add_widget(Button("Controls", 0, -0.45, command=self.request, extraArgs=["Controls"]))
		self.add_widget(Button("Configure Bots", 0, -0.65, command=self.request, extraArgs=["Robots"]))
		self.add_widget(Button("Exit", 0, -0.85, command=exit))
		self.scene.accept("escape", self.request, ["Visualizer"])

		self.scene.taskMgr.add(self.look, "followMouse")

	def exit(self) -> None:
		self.clear_widgets()
		self.scene.taskMgr.remove("followMouse")
		self.scene.ignore_all()

	def look(self, task: Task.Task) -> int:
		if self.scene.mouseWatcherNode.hasMouse():
			self.logo.look(self.scene.mouseWatcherNode.getMouseX(),
						   self.scene.mouseWatcherNode.getMouseY())
		return Task.cont


class Menu(State):
	def enter(self) -> None:
		self.add_widget(Slider("Movement Sensitivity", 0, 0.6, self.scene.controls.set_speed, int(self.scene.controls.speed*100)))
		self.add_widget(Slider("Zoom", 0, 0.2, self.scene.controls.set_zoom, int(self.scene.controls.zoom*100)))
		self.add_widget(Button("Back", 0, -0.25, command=self.request, extraArgs=["Visualizer"]))
		self.add_widget(Button("Controls", 0, -0.45, command=self.request, extraArgs=["Controls"]))
		self.add_widget(Button("Configure Bots", 0, -0.65, command=self.request, extraArgs=["Robots"]))
		self.add_widget(Button("Exit", 0, -0.85, command=exit))
		self.scene.accept("escape", self.request, ["Visualizer"])

	def exit(self) -> None:
		self.toRobots()
		self.scene.ignore_all()

	def toRobots(self) -> None:
		self.clear_widgets()
		self.scene.ignore_all()


class Controls(State):
	def enter(self) -> None:
		self.add_widget(Text("CONTROL SCHEME", 0, 0.8, 1))
		self.controls_text = Text(self.scene.controls.description, 0, 0.4, 1)
		self.add_widget(self.controls_text)
		options = [i for i in zip(self.scene.controls.list_names, self.scene.controls.list)]
		self.add_widget(Select(0, 0.6, options, self.on_switch, selected=self.scene.controls.selected))
		self.add_widget(Button("BACK", 0, -0.75, command=self.request,
			extraArgs=["Start" if self.scene.back_to_start else "Menu"]))

	def on_switch(self, scheme: str) -> None:
		self.scene.controls.request(scheme)
		self.controls_text.setText(self.scene.controls.description)

	def exit(self) -> None:
		self.clear_widgets()


class Robots(State):
	shift_down: bool
	frame: Options
	next: Button

	def enter(self, selected: list[str] = []) -> None:
		self.shift_down = False
		self.add_widget(Text("AVAILABLE ROBOTS", 0, 0.8, 1))
		self.add_widget(Text("Select the bots you want in your swarm from the list or", -0.18, 0.73, 0.5))
		self.add_widget(Button("add robots", 0.68, 0.73, scale=0.3, command=self.choose_file))
		self.frame = Options(0, 0.15, add_fn=self.choose_file, command=self.on_select)
		self.add_widget(self.frame)
		self.add_widget(Text("Press Ctrl+A to select all, Esc to deselect all", 0, -0.4, 0.4))
		self.add_widget(
			Text("Hold shift and click on a bot to select all bots between that one and the last one you selected ",
				0, -0.45, 0.4)
		)
		self.add_widget(Button("back", -0.85, -0.75, command=self.request,
			extraArgs=["Start" if self.scene.back_to_start else "Menu"]))
		self.next = Button("next", 0.85, -0.75, command=self.request, extraArgs=["Positions", self.frame.selected])
		self.next.disable()
		self.add_widget(self.next)
		self.scene.accept("control-a", self.frame.toggle_all, [True])
		self.scene.accept("escape", self.frame.toggle_all, [False])
		self.scene.accept("shift", lambda: setattr(self.frame, "shift_down", True))
		self.scene.accept("shift-up", lambda: setattr(self.frame, "shift_down", False))
		for bot in self.scene.bot_pool.values():
			self.frame.add_option(bot.hash, bot.texture_2d)
		for hash in selected:
			self.frame.toggle(hash, True)

	def exit(self, *args: Any) -> None:
		self.shift_down = False
		self.clear_widgets()
		self.frame.destroy()
		self.scene.ignore_all()

	def choose_file(self) -> None:
		file_path = fd.askopenfilename(initialdir = Settings.paths.config,title = "Select robots file",filetypes = (("json files","*.json"),("all files","*.*")))
		swarm = SwarmData(file_path)
		if not swarm.is_valid:
			self.scene.toaster.warn(swarm.error)
			return
		for i, bot_data in enumerate(swarm.swarm, 1):
			bot = Robot(bot_data)
			if not bot.is_valid:
				# todo create toast instead of printing to console
				self.scene.toaster.warn(f"Bot {i} in {swarm.file_name} is invalid: {bot.error}")
				continue
			if self.scene.add_robot(bot):
				self.frame.add_option(bot.hash, bot.texture_2d)

	def on_select(self, total_selected: int) -> None:
		"""The function that runs whenever a bot is selected or deselected"""
		if total_selected:
			self.next.enable()
			self.scene.accept("enter", self.request, ["Positions", self.frame.selected])
		else:
			self.scene.ignore("enter")
			self.next.disable()


class Positions(State):
	# todo load files
	active: DirectFrame
	swarm_size: int = 0
	terrain_width: int = 0
	terrain_length: int = 0
	positions: Optional[PositionData] = None
	min_dimensions: Tuple[int, int] = (0, 0)
	selected: list[str]
	max_bot_size: float = 0

	def enter(self, selected: set[str]) -> None:
		self.max_bot_size = self.scene.max_bot_size(selected)
		self.overlay = DirectGuiWidget(frameSize=(-10, 10, -10, 10), frameColor=Settings.theme.bg[:3]+(0.6,), sortOrder=10)
		self.overlay.hide()
		self.terrain_width = round(self.scene.terrain.get_sx()*2)
		self.terrain_length = round(self.scene.terrain.get_sy()*2)
		if not self.swarm_size: self.swarm_size = Settings.robots.default
		self.selected = list(selected)
		self.add_widget(Text("POSITIONS", 0, 0.8, 1))
		options = [("auto position", "auto"), ("load manually", "load")]
		self.add_widget(Select(0, 0.6, options, self.on_switch, selected=self.scene.layout))
		self.add_widget(Button("back", -0.85, -0.75, command=self.request, extraArgs=["Robots", selected]))
		self.render = Button("render", 0.85, -0.75, command=self.render_robots)
		self.add_widget(self.render)
		if self.scene.layout == "load": self.render.disable()
		self.auto = DirectFrame()
		self.add_widget(self.auto)
		Text("SWARM SIZE", -0.6, 0, 1).reparentTo(self.auto)
		NumEntry(self.swarm_size, Settings.robots.max, -0.63, -0.1, 1.52, 0.7, callback=self.set_swarm_size).reparentTo(self.auto)
		Text("TERRAIN DIMENSIONS", 0.6, 0, 1).reparentTo(self.auto)
		self.terrain_size = Text(f"{self.terrain_width} x {self.terrain_length}", 0.6, -0.1, 0.7)
		self.terrain_size.reparent_to(self.auto)

		self.load = DirectFrame()
		self.add_widget(self.load)
		Text("POSITIONS",            -0.6,  0,      1).reparentTo(self.load)
		Button("Choose File",        -0.6, -0.1, scale=0.5, command=self.choose_positions).reparent_to(self.load)
		Text("SWARM SIZE",		     -0.6, -0.22,    0.7).reparentTo(self.load)
		self.loaded_size = Text("0" if not self.positions else str(len(self.positions.positions)), -0.6, -0.3, 0.7, parent=self.load)
		Text("TERRAIN DIMENSIONS",    0.6,  0,      1).reparentTo(self.load)
		Button("Choose File",         0.6, -0.1, scale=0.5, command=self.choose_dimensions).reparentTo(self.load)
		Text("OR",                    0.6, -0.2,    0.7).reparentTo(self.load)
		self.width_entry = NumEntry(self.terrain_width, 0, 0.45, -0.3, 1.52, 0.7, self.set_terrain_width, parent=self.load)
		Text("x",                     0.6, -0.3,    0.7).reparentTo(self.load)
		self.length_entry = NumEntry(self.terrain_length, 0, 0.65, -0.3, 1.52, 0.7, self.set_terrain_length, parent=self.load)

		if self.scene.layout == "auto":
			self.load.hide()
			self.active = self.auto
		else:
			self.auto.hide()
			self.active = self.load

	def on_switch(self, mode: str) -> None:
		self.active.hide()
		self.active = getattr(self, mode)
		self.active.show()
		self.scene.layout = mode  # type: ignore
		if mode == "auto":
			self.render.enable()
		# elif self.terrain_width < self.min_dimensions[0] or self.terrain_length < self.min_dimensions[1]:
		else:
			self.render.disable()

	def exit(self, *args: Any) -> None:
		self.overlay.destroy()
		self.clear_widgets()

	def set_swarm_size(self, num: str) -> None:
		n = int(num)
		if n >= Settings.robots.threshold and n != self.swarm_size:
			self.scene.toaster.warn(f"Swarm sizes greater than {Settings.robots.threshold} may cause the program to slow down")
		self.swarm_size = n
		if n == 0:self.render.disable()
		else: self.render.enable()

		size = round((self.max_bot_size+Settings.robots.padding)*math.ceil(n**0.5))
		self.terrain_size["text"] = f"{size} x {size}"

	def set_terrain_width(self, num: str) -> None:
		self.terrain_width = int(num)
		if self.positions and self.terrain_width >= self.min_dimensions[0] and self.terrain_length >= self.min_dimensions[1]:
			self.render.enable()
		else:
			self.render.disable()

	def set_terrain_length(self, num: str) -> None:
		self.terrain_length = int(num)
		if self.positions and self.terrain_length >= self.min_dimensions[1] and self.terrain_width >= self.min_dimensions[0]:
			self.render.enable()
		else:
			self.render.disable()

	def choose_positions(self) -> None:
		file_path = fd.askopenfilename(initialdir = Settings.paths.config,title = "Select positions file",filetypes = (("text files","*.txt"),("all files","*.*")))
		positions = PositionData(file_path)
		if not positions.is_valid:
			self.scene.toaster.warn(positions.error)
			self.render.disable()
			return
		if positions.total > Settings.robots.max:
			self.scene.toaster.warn(f"{positions.total} positions loaded, which exceeds maximum of {Settings.robots.max}. Only the first {Settings.robots.max} will be used.")
		self.loaded_size["text"] = str(len(positions.positions))
		self.positions = positions
		bot_size = self.scene.max_bot_size(self.selected)
		if self.positions.fit_on(self.terrain_width, self.terrain_length, bot_size):
			self.render.enable()
		else:
			self.min_dimensions = self.positions.min_dimensions(bot_size)
			self.scene.toaster.warn(f"Terrain too small for given positions. Dimensions must be at least\n{self.min_dimensions[0]} x {self.min_dimensions[1]}")

	def choose_dimensions(self) -> None:
		file_path = fd.askopenfilename(initialdir = Settings.paths.config,title = "Select dimensions file",filetypes = (("text files","*.txt"),("all files","*.*")))
		terrain = TerrainData(file_path)
		if not terrain.is_valid:
			self.scene.toaster.warn(terrain.error)
			return
		self.width_entry.text = str(terrain.width)
		self.length_entry.text = str(terrain.length)

	def render_robots(self) -> None:
		self.overlay.show()
		self.scene.clear_instances()
		self.request("Visualizer")
		if self.scene.layout == "auto":
			self.scene.auto_position(self.selected, self.swarm_size)
		else:
			self.scene.scale_terrain(self.terrain_width, self.terrain_length)
			errors = not self.scene.manual_position(self.selected, self.positions.positions)
			if errors:
				self.request("Menu")
				self.scene.toaster.info("Switched to escape menu. You can return to the visualizer by pressing BACK once the collisions are resolved")

class Visualizer(State):
	def enter(self) -> None:
		self.fromMenu()

	def exit(self) -> None:
		self.toMenu()
		self.scene.terrain.remove_node()
		self.scene.clear_instances()

	def toMenu(self) -> None:
		self.scene.overlay.show()
		self.scene.controls.disable()
		self.scene.ignore_all()

	def fromMenu(self) -> None:
		self.scene.task_mgr.remove("orbit")
		self.scene.back_to_start = False
		self.scene.overlay.hide()
		self.scene.accept("escape", self.request, ["Menu"])
		self.scene.accept("b", self.scene.toggle_bounding_boxes)
		self.scene.controls.enable()

	def fromStart(self) -> None:
		self.fromMenu()