from __future__ import annotations
from abc import ABC
import math
from typing import Any, Callable, Optional, Tuple
from direct.gui.DirectGui import DirectFrame, DirectButton
from panda3d.core import TextNode, Texture, PNMImage, StringStream
from direct.gui.DirectGuiBase import DirectGuiWidget

import io
import threading

from PIL import Image
from PIL.ImageDraw import Draw

import roboviz.Settings as Settings
from roboviz.Utils import p3d_to_rgb

DEFAULT_DURATION = 10

class Toaster:
	__instance = None
	GAP = 0.02

	toasts: list[Toast]
	total_errors: int
	overlay: DirectGuiWidget
	tray: DirectFrame

	def __new__(cls, *args: Any, **kwargs: Any):
		if Toaster.__instance is None:
			Toaster.__instance = super().__new__(cls)
		return Toaster.__instance

	def __init__(self):
		self.toasts = []
		self.total_errors = 0
		self.overlay = DirectGuiWidget(sortOrder=99, frameSize=(-10, 10, -10, 10), frameColor=Settings.theme.toast.error[:3]+(0.4,))
		self.overlay.hide()
		self.tray = DirectFrame(
			sortOrder=100,
			pos=(0, 0, 1-Toaster.GAP)
		)

	def warn(self, message: str, duration: float = DEFAULT_DURATION):
		"""
		Display a warning popup with the given message for `duration` seconds. All warning popups
		can be dismissed by clicking on them.

		`message` - The warning message to display on the popup

		`duration` - The number of seconds to display the popup for, 0 to keep the popup. The
		default value is 10.
		"""
		w = Warn(self, message, len(self.toasts))
		self.toasts.append(w)
		for i in self.toasts[:-1]:
			i.set_z(i.get_z() - self.toasts[-1].height - Toaster.GAP)
		if duration != 0:
			t = threading.Timer(duration, self.toasts[-1].destroy)
			w.timer = t
			t.start()

	def error(self, message: str,
			buttons: list[str],
			callbacks: list[Callable],
			extraArgs: list[Tuple[Any, ...]]=[],
			dismissible: bool = False):
		"""
		Display an error popup with the given message and buttons. The buttons will be displayed
		in the order they are given in the `buttons` list. The `callbacks` list should contain
		one callback for each button, which will be called when the button is pressed.

		`message` - The error message to display on the popup

		`buttons` - A list of button labels to display on the popup

		`callbacks` - A list of callbacks to call when the corresponding button is pressed

		`extraArgs` - A list of tuples of extra arguments to pass to the corresponding callback

		`dismissible` - Whether the popup has a close button that can be used to dismiss it without
		taking an action. Generally if this is allowed the popup should be a warning instead of an
		error.
		"""
		self.total_errors += 1
		self.overlay.show()
		e = Error(self, message, len(self.toasts), buttons, callbacks, extraArgs, dismissible)
		self.toasts.append(e)
		for i in self.toasts[:-1]:
			i.set_z(i.get_z() - self.toasts[-1].height - Toaster.GAP)

	def info(self, message: str, duration: float = DEFAULT_DURATION):
		"""
		Display an information popup with the given message for `duration` seconds. All warning popups
		can be dismissed by clicking on them.

		`message` - The info message to display on the popup

		`duration` - The number of seconds to display the popup for, 0 to keep the popup. The
		default value is 10.
		"""
		inf = Info(self, message, len(self.toasts))
		self.toasts.append(inf)
		for i in self.toasts[:-1]:
			i.set_z(i.get_z() - self.toasts[-1].height - Toaster.GAP)
		if duration != 0:
			t = threading.Timer(duration, self.toasts[-1].destroy)
			inf.timer = t
			t.start()


class Toast(DirectFrame, ABC):
	WIDTH=1.8
	MIN_HEIGHT=0.2
	TEXT_OFFSET = (0.19, 0.075)
	BORDER_WIDTH = 0.015
	dismissible = True

	timer: Optional[threading.Timer]
	toaster: Toaster
	height: float
	position: int
	size: Tuple[float, float, float, float]

	def __init__(self, toaster: Toaster, message: str, position: int, extra_height: float = 0):
		self.timer = None
		self.toaster = toaster
		self.position = position
		node = TextNode("msg")
		node.set_text(message)
		node.set_text_scale(0.05)
		node.set_wordwrap(Toast.WIDTH - Toast.TEXT_OFFSET[0] - Toast.BORDER_WIDTH*2)
		node.set_text_color(*Settings.theme.gui.fg)

		self.height = max(Toast.MIN_HEIGHT, node.get_height()+Toast.TEXT_OFFSET[1])+extra_height
		self.size = (-Toast.WIDTH/2, Toast.WIDTH/2, -self.height, 0)

		super().__init__(frameSize=self.size, parent=toaster.tray)

		DirectButton(frameColor=Settings.theme.gui.bg,
			frameSize=[i-Toast.BORDER_WIDTH if i >= 0 else i+Toast.BORDER_WIDTH for i in self.size],
			relief=1, pressEffect=0, parent=self, command=(self.destroy if self.dismissible else None))
		DirectFrame(image=self.img, scale=0.07, pos=(self.size[0]+0.1, 0, self.size[2]/2), parent=self)

		text = self.attach_new_node(node)
		text.set_x(self.size[0]+Toast.TEXT_OFFSET[0])
		text.set_z(-0.075)

		self.initialiseoptions(Toast)

	def destroy(self):
		if self.timer: self.timer.cancel()
		self.toaster.toasts.pop(self.position)
		for i in range(self.position):
			self.toaster.toasts[i].set_z(self.toaster.toasts[i].get_z() + self.height + Toaster.GAP)

		for i in range(self.position, len(self.toaster.toasts)):
			self.toaster.toasts[i].position -= 1
		super().destroy()

	@property
	def img(self) -> Texture: ...


class Warn(Toast):
	__img: Optional[Texture] = None

	def __init__(self, parent: Toaster, message: str, position: int):
		super().__init__(parent, message, position)
		self["frameColor"] = Settings.theme.toast.warn
		self.initialiseoptions(Warn)

	@property
	def img(self) -> Texture:
		if not self.__img:
			img = Image.new("RGB", (100, 85), p3d_to_rgb(Settings.theme.gui.bg))
			canvas = Draw(img)
			canvas.polygon(
				((50, 5), (90, 80), (10, 80)),
				outline=p3d_to_rgb(Settings.theme.toast.warn),
				width=7
			)
			canvas.line((50, 35, 50, 60), fill=p3d_to_rgb(Settings.theme.gui.fg), width=7)
			canvas.line((50, 65, 50, 70), fill=p3d_to_rgb(Settings.theme.gui.fg), width=7)

			byte_arr = io.BytesIO()
			img.save(byte_arr, format='PNG')

			p = PNMImage()
			p.read(StringStream(byte_arr.getvalue()))
			tex = Texture()
			tex.load(p)
			self.__img = tex
		return self.__img

class Error(Toast):
	__img: Optional[Texture] = None
	BTN_OFFSET = 0.11
	BTN_WIDTH = 2
	dismissible = False
	BTN_POSITIONS = {1: (0,), 2:(-0.3, 0.3), 3: (-0.5, 0, 0.5)}

	def __init__(self,
			parent: Toaster,
			message: str,
			position: int,
			buttons: list[str],
			callbacks: list[Callable],
			extraArgs: list[Tuple[Any, ...]]=[],
			dismissible: bool = False):
		assert len(buttons) == len(callbacks), "The number of buttons and callbacks must be equal"
		if dismissible:
			buttons.append("Close")
			callbacks.append(lambda: None)
		rows = math.ceil(len(buttons)/3)
		super().__init__(parent, message, position, Error.BTN_OFFSET*rows)
		self["frameColor"] = Settings.theme.toast.error
		self.initialiseoptions(Error)

		for row in range(rows):
			pos = min(3, len(buttons)-row*3)
			positions = Error.BTN_POSITIONS[pos]
			for i in range(pos):
				index = i+row*3
				DirectButton(
					text=buttons[index].upper(),
					command= self.fn(callbacks[index]),
					extraArgs=extraArgs[index] if index < len(extraArgs) else [],
					pos=(positions[i], 0, self.size[2] + Error.BTN_OFFSET*(rows-row-0.5)),
					parent=self,
					scale=0.1,
					relief=1,
					text_fg=Settings.theme.gui.bg,
					frameColor=Settings.theme.gui.fg,
					text_scale=0.5,
					frameSize=(-Error.BTN_WIDTH, Error.BTN_WIDTH, -0.3, 0.6),
				)

	def destroy(self):
		super().destroy()
		self.toaster.total_errors -= 1
		if self.toaster.total_errors == 0: self.toaster.overlay.hide()

	def fn(self, callback: Callable):
		def _(*args: Any):
			self.destroy()
			return callback(*args)
		return _

	@property
	def img(self) -> Texture:
		if not self.__img:
			img = Image.new("RGB", (100, 100), p3d_to_rgb(Settings.theme.gui.bg))
			canvas = Draw(img)
			canvas.ellipse((9, 9, 89, 89), fill=p3d_to_rgb(Settings.theme.toast.error))
			canvas.line((34, 34, 64, 64), fill=p3d_to_rgb(Settings.theme.gui.bg), width=10)
			canvas.line((64, 34, 34, 64), fill=p3d_to_rgb(Settings.theme.gui.bg), width=10)

			byte_arr = io.BytesIO()
			img.save(byte_arr, format='PNG')

			p = PNMImage()
			p.read(StringStream(byte_arr.getvalue()))
			tex = Texture()
			tex.load(p)
			self.__img = tex
		return self.__img

class Info(Toast):
	__img: Optional[Texture] = None
	def __init__(self, parent: Toaster, message: str, position: int):
		super().__init__(parent, message, position)
		self["frameColor"] = Settings.theme.toast.info
		self.initialiseoptions(Info)

	@property
	def img(self) -> Texture:
		if not self.__img:
			img = Image.new("RGB", (100, 100), p3d_to_rgb(Settings.theme.gui.bg))
			canvas = Draw(img)
			canvas.ellipse((9, 9, 89, 89), outline=p3d_to_rgb(Settings.theme.toast.info), width=7)
			canvas.line((49, 25, 49, 35), fill=p3d_to_rgb(Settings.theme.toast.info), width=10)
			canvas.line((49, 45, 49, 73), fill=p3d_to_rgb(Settings.theme.toast.info), width=10)

			byte_arr = io.BytesIO()
			img.save(byte_arr, format='PNG')

			p = PNMImage()
			p.read(StringStream(byte_arr.getvalue()))
			tex = Texture()
			tex.load(p)
			self.__img = tex
		return self.__img